function dydt = CRISPRsim(t,y,Cas9_cassette,G,rev_G,indices,rev_indices,N,offtarget_energy,offtarget_sites, free)
% CRISPRSIM  Sets up the form of the ODEs which describe CRISPR-Cas9 action
% in chloroplasts
%   t: the time point
%   y: the state vector
%   Cas9_cassette: the number of "driver" cassettes in the chloroplast
%   G: the vector of free energy changes for Cas9 binding to viable sites
%   on the forward genome strand, and rev_G for the reverse strand
%   indices: the indices of forward sites to simulate changes in for G
%   vector, and rev_indices for rev_G
%   N: the length of the genome
%   offtarget_energy: combined Boltzmann energy of off-target sites with >1
%   nucleotide mismatch in the first 8 PAM-proximal nucleotides
%   offtarget_sites: number of off-target sites with >1 nucleotide mismatch
%   in the first 8 PAM-proximal nucleotides
%   free: a free variable for sensitivity analysis
%
% See also TESTS

m=0; % 1 to simulate using seconds, 0 for minutes

sel=1; % between 0 and 1, represents the strength of selection pressure (1 for "perfect")
forward_sites = length(indices);
rev_sites = length(rev_indices);
k_psaA = 70/475*60^(1-m); % gRNA transcription rate
deg_psaA = 0.129511805/60^(m+1); % gRNA degradation rate
k_atpA = 70/4000*60^(1-m); % Cas9 mRNA transcription rate
deg_atpA = 0.129511805/60^(m+1); % Cas9 mRNA degradation rate
k_Cas9 = 0.0057; % Cas9 translation rate
%k_Cas9 = 4.567e-6*60^(1-m); % Alternate potential Cas9 translation rate
k_B = 1.38064852e-23; % Boltzmann constant in m^2.kg.s^-2.K^-1
T = 298; % Temperature in Kelvin
divtime = 12; % C. reinhardtii cell division time in hours
mu = 1/(60^(m+1)*divtime); % Division rate per second
copy = 80; % Copy number
k_c = 0.0016; % Cas9 cleavage constant
k_d = 1; % Cas9 dissociation constant
k_I = 1*60^(1-m); % Cas9:gRNA inactive complex formation constant
k_f = 0.08*60^(1-m); % Cas9:gRNA inactive complex formation constant
D = 45*60^(1-m); % Cas9 diffusivity
lambda = 0.015; % Cas9 characteristic length in um
dydt = zeros(8 + forward_sites + rev_sites,1);
[~,cutsite] = min(G); % Sites are defined at 5' end of gRNA and Cas9 cuts 17bp upstream of that (3bp down of PAM)
deg_Cas9 = mu; % Cas9 "degradation by dilution" rate
deg_Cas9crRNA = min(deg_Cas9,deg_psaA); % Cas9:gRNA complex degradation rate
V = 25; % C. reinhardtii volume in um^3, Assuming chloroplast volume is half that of Chlamy cell
k_h = 0.43/60^m; % Homologus recombination rate constant

% k_h = free; % This is an example of how to perform a sensitivity analysis on k_h (if this line is uncommented): see TESTS

dydt(1) = k_psaA*Cas9_cassette - deg_psaA*y(1) - k_f*y(1)*y(3); % gRNA ODE
dydt(2) = k_atpA*Cas9_cassette - deg_atpA*y(2); % Cas9 mRNA ODE
dydt(3) = k_Cas9*y(2) - deg_Cas9*y(3) - k_f*y(1)*y(3); % Cas9 ODE
dydt(4) = k_f*y(1)*y(3) - deg_Cas9crRNA*y(4) - k_I*y(4); % Cas9:gRNA inactive complex ODE

r_RW = 6*D*lambda*y(5)/V; % Rate of random walk (1/tau)

r_c = zeros(1,forward_sites);
r_c_rev = zeros(1,rev_sites);

denom = 2*N; % Initialise denominator of Boltzmann function

% Add contribution of each site to the denominator of Boltzmann function
for j = 1:min(length(G),length(rev_G))
    
    a = find(indices==j);
    b = find(rev_indices==j);
    
    if a > 0
        if b > 0 % For sites where both forward and reverse genome locations are targets
            denom = denom + y(5+a)*exp(-G(j)/(k_B*T))+ y(5+forward_sites+b)*exp(-rev_G(j)/(k_B*T));
        else % Forward genome location is target but reverse isn't
            denom = denom + y(5+a)*exp(-G(j)/(k_B*T)) + copy*exp(-rev_G(j)/(k_B*T));
        end
    elseif b < 0 % Reverse genome location is target but forward isn't
        denom = denom + copy*exp(-G(j)/(k_B*T)) + y(5+forward_sites+b)*exp(-rev_G(j)/(k_B*T));
    else % Neither are targets
        denom = denom + copy*exp(-G(j)/(k_B*T)) + copy*exp(-rev_G(j)/(k_B*T));
    end
end

denom = denom + offtarget_energy*y(end-2); % Add contibution of all off-target sites which have been combined into one ODE

% Continue adding to the denominator until contributions of all sites have
% been considered
if length(G) > length(rev_G)
    for j=length(rev_G)+1:length(G)
        a = find(indices==j);
        if a > 0
            denom = denom + y(5+a)*exp(-G(j)/(k_B*T));
        else
            denom = denom + copy*exp(-G(j)/(k_B*T));
        end
    end
elseif length(G) < length(rev_G)
    for j=length(G)+1:length(rev_G)
        a = find(rev_indices==j);
        if a > 0
            denom = denom + y(5+forward_sites+a)*exp(-rev_G(j)/(k_B*T));
        else
            denom = denom + copy*exp(-rev_G(j)/(k_B*T));
        end
    end
end

% Calculate cleavage rates on forward strand
for j = 1:forward_sites
    r_c(j) = k_c/(k_c+k_d)*y(5+j)*r_RW*exp(-G(j)/(k_B*T)) / denom;
end

% Calculate cleavage rates on reverse strand
for j = 1:rev_sites
    r_c_rev(j) = k_c/(k_c+k_d)*y(5+length(indices)+j)*r_RW*exp(-rev_G(j)/(k_B*T)) / denom;
end

% ODE for Cas9:gRNA active complex
dydt(5) = k_I*y(4) - deg_Cas9crRNA*y(5) - sum(r_c) - sum(r_c_rev) - k_c/(k_c+k_d)*y(end-2)*r_RW*offtarget_energy/(denom*offtarget_sites); % Change in active Cas9-gRNA complexes

% Set up ODEs for target sites on forward strand (NOTE THAT THIS CODE
% ASSUMES THE ON-TARGET SITE IS ON THE FORWARD STRAND)
for j = 1:forward_sites
    if j == cutsite
        dydt(5+j) = (1-sel)*k_h*(y(end-1)-y(5+j)) - r_c(j);
    else
        dydt(5+j) = k_h*(copy-y(5+j)) - r_c(j);
    end
end

% Set up ODEs for target sites on reverse strand
for j = 1:rev_sites
    dydt(5+forward_sites+j) = k_h*(copy-y(5+forward_sites+j)) - r_c_rev(j);
end

% ODE for average off-target sites with >1 mismatch in its first 8
% nucleotides
dydt(end-2) = k_h*(copy-y(end-2)) - k_c/(k_c+k_d)*y(end-2)*r_RW*offtarget_energy/(denom*offtarget_sites);

dydt(end-1) = -(y(end-1)-y(5+cutsite))*k_h*y(end)*sel; % ODE for untransformed sites (both bound and unbound)
dydt(end) = sel*k_h*(y(end-1)-y(5+cutsite))*y(end); % ODE for transformed sites

end