% CAS9 simulates the process of homoplasmic, Cas9-aided chloroplast
% transformation. It outputs an expected timescale for a homoplasmic
% transformation, from the point of the insertion of the cassettes, to the
% display window

load InvitroModel; % Load energy exchange and weight data fitted from an in vitro model provided by Prof. Howard Salis, Penn State University

avo = 6.0221409e+23; % Avogadro's number
k_B = 1.38064852e-23; % Boltzmann constant in m^2.kg.s^-2.K^-1
T = 298; % Temperature in Kelvin
copy = 80; % Copy number

% gRNA corresponding to the site on the genome to target
gRNA = 'AACGCGTTTATCTTAACGGA';

% Read genome from text file
fileID = fopen('chlamy genome.txt','r');
formatSpec = '%s';
genome = fscanf(fileID,formatSpec);

% Initialise "worst case scenario" numbers of cassettes inserted into
% chloroplast by co-transformation
gRNA_cassette = 1;
Cas9_cassette = 1;

N = length(genome); % Genome length
G = zeros(1,N); bind_energy = zeros(1,length(G));
G_rev = zeros(size(G)); rev_bind_energy = zeros(1,length(G));

genome = [genome genome(1:23)]; % Circularly extend genome

% Generate reverse strand
rev_genome = fliplr(genome);
rev_genome = strrep(rev_genome,'A','P');
rev_genome = strrep(rev_genome,'T','A');
rev_genome = strrep(rev_genome,'P','T');
rev_genome = strrep(rev_genome,'G','P');
rev_genome = strrep(rev_genome,'C','G');
rev_genome = strrep(rev_genome,'P','C');

% PAM free energy lookup table
PAM_table = {'AAA','ATA','ACA','AGA'; 'AAT','ATT','ACT','AGT';'AAC','ATC','ACC','AGC';'AAG','ATG','ACG','AGG';'TAA','TTA','TCA','TGA';'TAT','TTT','TCT','TGT';'TAC','TTC','TCC','TGC';'TAG','TTG','TCG','TGG';'CAA','CTA','CCA','CGA';'CAT','CTT','CCT','CGT';'CAC','CTC','CCC','CGC';'CAG','CTG','CCG','CGG';'GAA','GTA','GCA','GGA';'GAT','GTT','GCT','GGT';'GAC','GTC','GCC','GGC';'GAG','GTG','GCG','GGG'};
G_PAM = [0,0,0,-7.5; -6.6,-6.8, 0,-7.7; 0,-6,0,-7.7; -6.6,0,0,-7.7; 0,0,0,0; -6.8,-7.2,0,0; 0,0,0,0; -6.8,-6.4,0,-7.4; 0,0,0,0; 0,0,0,0; 0,0,0,0; 0,0,0,-7.7; -6.8,-7,0,-9.4; -6.9,-7.5,-6.7,-9.6; -6.8,0,0,-9.7; -6.9,-7.3,-6.5,-9.5];

DNA_lookup_table = 'ATCG';
weights=w1;

% --------------------------------------
% Calculate average mismatch free energy
sites = 0;
tot_G = 0;

% Go through genome site by site
for j = 1:N
    % Site is defined at 3' end so 3-nucleotide PAM is 21 nucleotides down
    PAM = genome(j+21:j+23);
    
    for k = 0:19 % along length of 20 nucleotide gRNA
        % a mismatch in 1 nucleotide = a mismatch in 2 duplexes for most,
        % or in 1 duplex for sites at beginning and end
        if genome(j+k) ~= gRNA(k+1) % If there is a mismatch
            if k > 0 % Add free energy from 3' duplex
                
                % find indices for decNN
                DNA_index1 = find(DNA_lookup_table == genome(j+k-1))+4*(find(DNA_lookup_table == (genome(j+k)))-1); 
                RNA_index1 = find(DNA_lookup_table == gRNA(k+1-1))+4*(find(DNA_lookup_table == (gRNA(k+1)))-1);
                
                dG1 = decNN(RNA_index1,DNA_index1); % Find mismatch energy in decNN
                
                if (abs(dG1 - 0.000015) > 1e-6) % Unknown duplex energies were set to 0.000015 so neglect the contribution of that site if the value is encountered
                    tot_G = tot_G + dG1; % Add to total mismatch energy calculated
                    sites = sites + 1; % Add to number of mismatch sites
                end
                
            end
            if k < 19 % Add free energy from 5' duplex
                DNA_index2 = find(DNA_lookup_table == genome(j+k))+4*(find(DNA_lookup_table == (genome(j+k+1)))-1);
                RNA_index2 = find(DNA_lookup_table == gRNA(k+1))+4*(find(DNA_lookup_table == (gRNA(k+2)))-1);
                dG2 = decNN(RNA_index2,DNA_index2);
                if (abs(dG2 - 0.000015) > 1e-6)
                    tot_G = tot_G + dG2;
                    sites = sites + 1;
                end
            end
        end
    end
    
    bind_energy(j) = G_PAM(strcmp(PAM_table,PAM)); % Save energy from PAMs binding
    
    % Same for reverse strand    
    PAM = rev_genome(j+21:j+23);
    
    for k = 0:19
        if rev_genome(j+k) ~= gRNA(k+1)
            if k > 0
                DNA_index1 = find(DNA_lookup_table == rev_genome(j+k-1))+4*(find(DNA_lookup_table == (rev_genome(j+k)))-1);
                RNA_index1 = find(DNA_lookup_table == gRNA(k+1-1))+4*(find(DNA_lookup_table == (gRNA(k+1)))-1);
                dG1 = decNN(RNA_index1,DNA_index1);
                if (abs(dG1 - 0.000015) < 1e-6)
                    sites = sites - 1;
                else
                    tot_G = tot_G + dG1;
                    sites = sites + 1;
                end
            end
            if k < 19
                DNA_index2 = find(DNA_lookup_table == rev_genome(j+k))+4*(find(DNA_lookup_table == (rev_genome(j+k+1)))-1);
                RNA_index2 = find(DNA_lookup_table == gRNA(k+1))+4*(find(DNA_lookup_table == (gRNA(k+2)))-1);
                dG2 = decNN(RNA_index2,DNA_index2);
                if (abs(dG2 - 0.000015) < 1e-6)
                    sites = sites - 1;
                else
                    tot_G = tot_G + dG2;
                    sites = sites + 1;
                end
            end
        end
    end
    
    rev_bind_energy(j) = G_PAM(strcmp(PAM_table,PAM));
end

meanNN = tot_G/sites; % Average mismatch free energy change
disp(['Average mismatch free energy: ', num2str(meanNN), ' kcal/mol'])

% -------------------------------------

% Vectors for number of mismatches in first 8 PAM-proximal nucleotides at
% each site
first_8 = zeros(1,length(G));
rev_first_8 = zeros(1,length(G_rev));

% Calculate energy exchange involved in Cas9 binding to each viable site
for j = 1:N
    
    d_G_exchange = 0;
    rev_d_G_exchange = 0;
    
    % Move along length of gRNA looking for mismatches
    for k = 0:19
        if genome(j+k) ~= gRNA(k+1)
            if k == 0 || k == 19 % Only 1 duplex mismatch if at beginning or end of gRNA
                d_G_exchange = d_G_exchange + meanNN*weights(20-k);
            else % Otherwise 2
                d_G_exchange = d_G_exchange + 2*meanNN*weights(20-k);
            end
            if k > 11 % Add to number of mismatches in the first 8 PAM-proximal nucleotides
                first_8(j) = first_8(j) + 1;
            end
        end
        % Same on reverse strand
        if rev_genome(j+k) ~= gRNA(k+1)
            if k == 0 || k == 19
                rev_d_G_exchange = rev_d_G_exchange + meanNN*weights(20-k);
            else
                rev_d_G_exchange = rev_d_G_exchange + 2*meanNN*weights(20-k);
            end
            if k > 11
                rev_first_8(j) = rev_first_8(j) + 1;
            end
        end
    end
    
    % Save final energy exchanges involved in Cas9 binding to all sites
    G(j) = bind_energy(j)+d_G_exchange;
    G_rev(j) = rev_bind_energy(j)+rev_d_G_exchange;
end

% convert from kcal/mol to J
G=G*4184/avo;
G_rev=G_rev*4184/avo;

index_calc = [G; bind_energy; first_8];
% Indices (in G array) of sites with negative binding energy, a canonical
% PAM, and <= 1 mismatch in first 8 nucleotides (i.e. those which will be
% simulated using their own ODEs)
indices = find(index_calc(1,:) < 0 & index_calc(2,:) < 0 & index_calc(3,:) <= 1);
% Indices (in G array) of sites with negative binding energy, a canonical
% PAM, and > 1 mismatch in first 8 nucleotides (i.e. those which will be
% averaged into a single ODE)
offtarget_indices = find(index_calc(1,:) < 0 & index_calc(2,:) < 0 & index_calc(3,:) > 1);

% Same for reverse strand
index_calc = [G_rev; rev_bind_energy; rev_first_8];
rev_indices = find(index_calc(1,:) < 0 & index_calc(2,:) < 0 & index_calc(3,:) <= 1);
rev_offtarget_indices = find(index_calc(1,:) < 10000 & index_calc(2,:) < 0 & index_calc(3,:) > 1);

% Initialise Boltzmann energy for all sites with >= 1 mismatch in first 8
% nucleotides
offtarget_energy=0;

% Add Boltzmann contribution for all >= 1 mismatch sites on forward strand
for j=1:length(offtarget_indices)
    offtarget_energy = offtarget_energy + exp(-G(offtarget_indices(j))/(k_B*T));
end

% Add Boltzmann contribution for all >= 1 mismatch sites on forward strand
for j=1:length(rev_offtarget_indices)
    offtarget_energy = offtarget_energy + exp(-G_rev(rev_offtarget_indices(j))/(k_B*T));
end

% Trim G vector to only include target sites with negative free energy and
% <= 1 mismatch in first 8 nucleotides
G=G(indices);
G_rev=G_rev(rev_indices);

% Initialise state vector
% y(1) = gRNA; y(2) = Cas9 mRNA; y(3) = Cas9;
% y(4) = inactive complex; y(5) = active complex
% y(6:end-3) = target sites with <=1 mismatch in first 8 PAM-proximal nucleotides (including on-target site)
% y(end-2) = average target site with > 1 mismatch in first 8 nucleotides
% y(end-1) = number of untransformed on-target sites (both bound and unbound)
% y(end) = copies of gene of interest
initial_y = zeros(1,length(indices)+length(rev_indices)+8);
initial_y(6:end-1) = copy;
initial_y(end) = gRNA_cassette;

free = 0; % Free parameter to vary for sensitivity analysis

opts = odeset('Events',@stopevent); % Set ODE to end when homoplasmy is achieved

% Simulate ODE system over a full year, stopping when homoplasmy is
% achieved
[t,y] = ode15s(@CRISPRsim,[0 52*7*24*60],initial_y,opts,Cas9_cassette,G,G_rev,indices,rev_indices,N,offtarget_energy,length(rev_offtarget_indices)+length(offtarget_indices), free);

disp(['Time to homoplasmy: ', num2str(t(end)/60), ' hours'])